/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors, inconsistencies, and inaccuracies with real-world hardware
implementations of the IA64 instruction set. If you build a mission-
critical system based only on this simulator and something goes wrong,
it's your own damned fault.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@glue.umd.edu  12/5/99
--------------------------------------------------------------------*/
#include "ia64.h"
#include "execute.h"


extern char ldhint[4][8];
extern char sthint[4][8];

/*---------------------------------------------------------------------*/
uint32 m_ldir(qword *instr) {

   uint32 rd, rs1, rs2, opcode, status;
   uint64 addr;

   rd = BITS(instr[0],6,12);
   rs1 = BITS(instr[0],13,19);
   rs2 = BITS(instr[0],20,26);
   opcode = BITS(instr[0],30,36)<<1 | BITS(instr[0],27,27);

   if(rs2) check_reg(rs2 | ALU_REG);

   switch(opcode) {

      case 0*2:
         printf("ld1%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         if(r[rs1].nat) return NAT_CONSUMPTION;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,1)) )
            return status;
         r[rd].nat = 0;
         return IA_OK;

      case 1*2:
         printf("ld2%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         if(r[rs1].nat) return NAT_CONSUMPTION;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,2)) )
            return status;
         r[rd].nat = 0;
         return IA_OK;

      case 2*2:
         printf("ld4%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         if(r[rs1].nat) return NAT_CONSUMPTION;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,4)) )
            return status;
         r[rd].nat = 0;
         return IA_OK;

      case 3*2:
         printf("ld8%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         if(r[rs1].nat) return NAT_CONSUMPTION;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,8)) )
            return status;
         r[rd].nat = 0;
         return IA_OK;

      case 4*2:
         printf("ld1.s%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         r[rd].nat = r[rs2].nat | (load_mem((uint8 *)(&r[rd].dat.q),&addr,1)!=IA_OK);
         return IA_OK;

      case 5*2:
         printf("ld2.s%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         r[rd].nat = r[rs2].nat | (load_mem((uint8 *)(&r[rd].dat.q),&addr,2)!=IA_OK);
         return IA_OK;

      case 6*2:
         printf("ld4.s%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         r[rd].nat = r[rs2].nat | (load_mem((uint8 *)(&r[rd].dat.q),&addr,4)!=IA_OK);
         return IA_OK;

      case 7*2:
         printf("ld8.s%s r%d = [r%d]\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         addr = r[rs2].dat.q;
         r[rd].nat = r[rs2].nat | (load_mem((uint8 *)(&r[rd].dat.q),&addr,8)!=IA_OK);
         return IA_OK;

      case 1*128+0*2:
         printf("ld1%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         if(r[rs1].nat || r[rs2].nat) return NAT_CONSUMPTION;
         addr = r[rs2].dat.q;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,1)) )
            return status;
         r[rs2].dat.q += r[rs1].dat.q;
         r[rd].nat = 0;
         return IA_OK;

      case 1*128+1*2:
         printf("ld2%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         if(r[rs1].nat || r[rs2].nat) return NAT_CONSUMPTION;
         addr = r[rs2].dat.q;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,2)) )
            return status;
         r[rs2].dat.q += r[rs1].dat.q;
         r[rd].nat = 0;
         return IA_OK;

      case 1*128+2*2:
         printf("ld4%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         if(r[rs1].nat || r[rs2].nat) return NAT_CONSUMPTION;
         addr = r[rs2].dat.q;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,4)) )
            return status;
         r[rs2].dat.q += r[rs1].dat.q;
         r[rd].nat = 0;
         return IA_OK;

      case 1*128+3*2:
         printf("ld8%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         if(r[rs1].nat || r[rs2].nat) return NAT_CONSUMPTION;
         addr = r[rs2].dat.q;
         if( (status=load_mem((uint8 *)(&r[rd].dat.q),&addr,8)) )
            return status;
         r[rs2].dat.q += r[rs1].dat.q;
         r[rd].nat = 0;
         return IA_OK;

      case 1*128+4*2:
         printf("ld1.s%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         addr = r[rs2].dat.q;
         r[rs2].nat |= r[rs1].nat;
         r[rd].nat = r[rs2].nat |
              (load_mem((uint8 *)(&r[rd].dat.q),&addr,1)!=IA_OK);
         r[rs2].dat.q += r[rs1].dat.q;
         return IA_OK;

      case 1*128+5*2:
         printf("ld2.s%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         addr = r[rs2].dat.q;
         r[rs2].nat |= r[rs1].nat;
         r[rd].nat = r[rs2].nat |
              (load_mem((uint8 *)(&r[rd].dat.q),&addr,2)!=IA_OK);
         return IA_OK;

      case 1*128+6*2:
         printf("ld4.s%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         addr = r[rs2].dat.q;
         r[rs2].nat |= r[rs1].nat;
         r[rd].nat = r[rs2].nat |
              (load_mem((uint8 *)(&r[rd].dat.q),&addr,4)!=IA_OK);
         return IA_OK;

      case 1*128+7*2:
         printf("ld8.s%s r%d = [r%d], r%d\n", ldhint[BITS(instr[0],28,29)],
                                      rd,rs2,rs1);
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         add_hazard(rs2 | ALU_REG);
         if(rs1) check_reg(rs1 | ALU_REG);
         if(rd==rs2) return ILLEGAL_INSTR;
         addr = r[rs2].dat.q;
         r[rs2].nat |= r[rs1].nat;
         r[rd].nat = r[rs2].nat |
              (load_mem((uint8 *)(&r[rd].dat.q),&addr,8)!=IA_OK);
         return IA_OK;
                                      
      case 0x30*2:
         printf("st1%s [r%d] = r%d\n", sthint[BITS(instr[0],28,29)],
                                      rs2,rs1);
         check_reg(rs1 | ALU_REG);
         addr = r[rs2].dat.q;
         if( (status=store_mem((uint8 *)(&r[rs1].dat.q),&addr,1)) )
            return status;
         return IA_OK;

      case 0x31*2:
         printf("st2%s [r%d] = r%d\n", sthint[BITS(instr[0],28,29)],
                                      rs2,rs1);
         check_reg(rs1 | ALU_REG);
         addr = r[rs2].dat.q;
         if( (status=store_mem((uint8 *)(&r[rs1].dat.q),&addr,2)) )
            return status;
         return IA_OK;

      case 0x32*2:
         printf("st4%s [r%d] = r%d\n", sthint[BITS(instr[0],28,29)],
                                      rs2,rs1);
         check_reg(rs1 | ALU_REG);
         addr = r[rs2].dat.q;
         if( (status=store_mem((uint8 *)(&r[rs1].dat.q),&addr,4)) )
            return status;
         return IA_OK;

      case 0x33*2:
         printf("st8%s [r%d] = r%d\n", sthint[BITS(instr[0],28,29)],
                                      rs2,rs1);
         check_reg(rs1 | ALU_REG);
         addr = r[rs2].dat.q;
         if( (status=store_mem((uint8 *)(&r[rs1].dat.q),&addr,8)) )
            return status;
         return IA_OK;

      default:
         return UNIMPLEMENTED;
   }
}

/*---------------------------------------------------------------------*/
uint32 m_mgmt(qword *instr) {

   uint32 sof, soi, sol, sor, rd, opcode;
   uint64 imm;

   opcode = BITS(instr[0],33,40);

   switch(opcode) {

      case 1*16+6:
         rd = BITS(instr[0],6,12);
         sof = BITS(instr[0],13,19);
         sol = BITS(instr[0],20,26);
         sor = BITS(instr[0],27,30) << 3;
         soi = cfm.q & 127,
         printf("alloc r%d = ar.pfs, %d,%d,%d,%d\n", rd,
                                    soi, sof - soi,
                                    sof - sol, sor );
         check_reg(rd | ALU_REG);
         add_hazard(rd | ALU_REG);
         if( sof>=soi && sof>=sol && sol>=sor ) {
            stop();
            r[rd].dat.q = ar[AR_PFS].q;
            cfm.q = BITS(instr[0],13,30);
            cfm.q |= 0x1fff<<28 | (sor-1)<<18;
         }
         else {
            return ILLEGAL_INSTR;
         }
         return IA_OK;

      case 0:
         imm = -(int64)BITS(instr[0],36,36)<<20 |
                       BITS(instr[0],6,25);
         if( BITS(instr[0],27,40)==1 ) {
            printf("nop.m %d\n", (uint32)imm);
            return IA_OK;
         }

      default:
         return UNIMPLEMENTED;
   }            
}

/*---------------------------------------------------------------------*/
uint32 m_ldii(qword *instr) { return UNIMPLEMENTED; }
uint32 m_ldfr(qword *instr) { return UNIMPLEMENTED; }
uint32 m_ldfi(qword *instr) { return UNIMPLEMENTED; }
