/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors, inconsistencies, and inaccuracies with real-world hardware
implementations of the IA64 instruction set. If you build a mission-
critical system based only on this simulator and something goes wrong,
it's your own damned fault.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@glue.umd.edu  12/5/99
--------------------------------------------------------------------*/
/* The simulator code */

#include "ia64.h"
#include "ia64asm.h"

#define MIN(a,b) ((a)<(b)? (a):(b))

char *nextchar, tokens[5] = " =";
char cmd[100];
uint64 breaks[8];
uint32 currslot;
uint64 currmem;

#define SLOT0 0x1
#define SLOT1 0x2
#define SLOT2 0x4

/*-----------------------------------------------------------------*/
void help(void) {
   printf("\n");
   printf("List of commands ( '{}' is optional )\n\n");
   printf("target = value   : assign 'value' (number or register) to\n");
   printf("                       'target' (memory or register). Spec-\n");
   printf("                       ifying ip as target resets the current\n");
   printf("                       instruction slot to slot 0\n");
   printf("B addr           : set a breakpoint at (bundle holding) 'addr'\n");
   printf("C file base      : assemble 'file' and load at 'base'\n");
   printf("G                : execute until breakpoint is hit\n");
   printf("I                : execute next instruction (not bundle)\n");
   printf("L file base      : loads binary file 'file' starting at 'base'\n");
   printf("P                : display predicate registers\n");
   printf("Q                : quit\n");
   printf("R start stop     : display like register types in current frame\n");
   printf("                       from index 'start' to 'stop'\n");
   printf("S file addr size : stores 'size' bytes to 'file' from 'addr'\n");
   printf("X {base}         : examine memory {at address 'base'}\n");
   printf("                       (constants can be decimal or '0x'-hex)\n");
   printf("                       Default is page after last use of 'X'\n");
}

/*-----------------------------------------------------------------*/
void pred(void) {

   int32 i;

   printf("predicate registers:\n");
printf("   6         5         4         3         2         1         0\n");
printf("3210987654321098765432109876543210987654321098765432109876543210\n");
printf("----------------------------------------------------------------\n");

   for(i=63; i>=0; i--) {
      if( (p[i]&1) )
         printf("1");
      else
         printf("0");
   }

   printf("\n");
}

/*-----------------------------------------------------------------*/
void assign(void) {

   char *first, *second, *junk;
   uint32 size1, size2, offset;
   uint64 i1, i2;

   first = strtok(cmd, tokens);
   second = strtok(NULL, tokens);
   if(!first || !second) {
      printf("1syntax error\n");
      return;
   }
   
   if(!isdigit(*first)) {           /*-------- PARSE FIRST ARGUMENT ------*/

      if(first[0]=='i' && first[1]=='p') {
         first = (char *)(&ip);
         size1 = sizeof(uint64);
         currslot = SLOT0;
      }
      else {
         i1 = strtoll(first+1, &junk, 0);
         if(*junk) {
            printf("3syntax error\n");
            return;
         }

         switch(*first) {
            case 'r':
               if( i1 > ( (cfm.q&127) + 32) || i1<0) {
                  printf("register outside frame\n");
                  return;
               }
               first = (char *)(r + i1);
               size1 = sizeof(alu_reg);
               break;
   
            case 'b':
               if( i1 > 7 || i1<0) {
                  printf("4syntax error\n");
                  return;
               }
               first = (char *)(b + i1);
               size1 = sizeof(uint64);
               break;
   
            case 'p':
               if( i1 > 63 || i1<0) {
                  printf("5syntax error\n");
                  return;
               }
               first = (char *)(p + i1);
               size1 = 1;
               break;
   
            default:
               printf("not implemented\n"); return;
         }
      }
   }
   else {
      i1 = strtoll(first, &junk, 0);
      if(*junk) {
         printf("6syntax error\n");
         return;
      }
      if(i1 > MEMSIZE) {
         printf("address out of range\n");
         return;
      }
      first = mem + (i1 & ~7);
      size1 = sizeof(uint64);
   }
                                   /*-------- PARSE SECOND ARGUMENT ------*/

   if(!isdigit(*second) && second[0]!='-') { 

      if(second[0]=='i' && second[1]=='p') {
         second = (char *)(&ip);
         size2 = sizeof(uint64);
      }
      else {
         i2 = strtoll(second+1, &junk, 0);
         if(*junk) {
            printf("7syntax error\n");
            return;
         }

         switch(*second) {
            case 'r':
               if( i2 > ( (cfm.q&127) + 32) || i2<0) {
                  printf("register outside frame\n");
                  return;
               }
               second = (char *)(r + i2);
               size2 = sizeof(uint64);
               break;
   
            case 'b':
               if( i2 > 7 || i2<0) {
                  printf("8syntax error\n");
                  return;
               }
               second = (char *)(b + i2);
               size2 = sizeof(uint64);
               break;
   
            case 'p':
               if( i2 > 63 || i2<0) {
                  printf("9syntax error\n");
                  return;
               }
               second = (char *)(p + i2);
               size2 = 1;
               break;
   
            default:
               printf("not implemented\n");
               return;
         }
      }
   }
   else {
      i2 = strtoll(second, &junk, 0);
      if(*junk) {
         printf("Asyntax error\n");
         return;
      }
      second = (char *)(&i2); size2 = sizeof(uint64);
   }

   memcpy(first, second, MIN(size1, size2));      /* do the assignment */
}

/*-----------------------------------------------------------------*/
void showreg(void) {

   char *first, *second, *junk;
   uint32 i1, i2, i;

   first = strtok(NULL, tokens);
   second = strtok(NULL, tokens);
   if(!first || !second || *first!=*second) {
      printf("1syntax error\n");
      return;
   }
   
   i1 = strtoul(first+1, &junk, 10);
   if(*junk) {
      printf("2syntax error\n");
      return;
   }
   i2 = strtoul(second+1, &junk, 10);
   if(*junk) {
      printf("3syntax error\n");
      return;
   }
   if(i1>i2) {
      printf("4syntax error\n");
      return;
   }

   switch( *first ) {

      case 'r':
         if(i2 >= ((cfm.q&127)+32) ) {
            printf("register outside frame\n");
            return;
         }
         printf("\n");
         for(i=0; i<(i2-i1+1); i++) {
            printf("r%d: 0x%08x%08x NAT: %d    ", i1+i,
                      HI(r[i1+i].dat), LO(r[i1+i].dat), r[i1+i].nat);
            if(i&1) printf("\n");
         }

         printf("\n\nip: 0x%08x%08x      ", HI(ip), LO(ip));
         printf("cfm: 0x%08x%08x\n", HI(cfm), LO(cfm));
         printf("ar.lc: 0x%08x%08x   ", HI(ar[AR_LC]), LO(ar[AR_LC]));
         printf("ar.ec: 0x%08x%08x\n", HI(ar[AR_EC]), LO(ar[AR_EC]));
         return;

      case 'b':
         if(i2 > 7  ) {
            printf("index out of range\n");
            return;
         }
         printf("\n");
         for(i=0; i<(i2-i1+1); i++) {
            printf("b%d: 0x%08x%08x    ", i1+i, HI(b[i1+i]),
                      LO(b[i1+i]) );
            if((i%3+1)==3) printf("\n");
         }
         printf("\n");
         return;

      default:
         printf("not supported\n");
   }
}

/*-----------------------------------------------------------------*/
void showmem(void) {

   bundle vliw;
   char *junk;
   uint8 *base;
   uint64 temp;
   uint32 i, j, mask;

   base = strtok(NULL, tokens);
   if(base) {
      temp = strtoll(base, &junk, 0);
      if(*junk) {
         printf("2syntax error\n");
         return;
      }
      currmem = temp & ~15;
   }

   for(i=0; i<64 && currmem<MEMSIZE; i+=16) {

      memcpy((char *)(&vliw), mem+currmem, sizeof(bundle));

      base = mem+currmem;

      for(j=0,mask=1; j<4; j++) {
         decode_instruction( &vliw, mask, cmd);
         printf("   %08x: 0x%02x%02x%02x%02x ", (uint32)currmem + 4*j,
                   base[3],base[2],base[1],base[0] );

         if(j<3) {
            if(ip.q==currmem && currslot==mask)
               printf("-> ");
            else
               printf("   ");
            printf("%s", cmd);
            mask <<= 1;
         }

         base += 4;
         printf("\n");
      }

      currmem += 16;
   }      
}

/*-----------------------------------------------------------------*/
void compile(void)  {

   FILE *infile;
   char *first, *second, *junk;
   uint64 base;
   uint32 status;
   bundle vliw;

   first = strtok(NULL, tokens);
   second = strtok(NULL, tokens);
   if(!first || !second) {
      printf("1syntax error\n");
      return;
   }
   
   base = strtoll(second, &junk, 0) & ~15;
   if(*junk) {
      printf("3syntax error\n");
      return;
   }

   infile = fopen(first, "r");
   if(!f) {
      printf("couldn't open '%s'\n", first);
      return;
   }

   while(base<MEMSIZE) {
      status = bundle_from_file( &vliw, infile );
      if(status!=IA_OK && status!=DRAIN) break;
      memcpy(mem+base, (char *)(&vliw), sizeof(bundle));
      base += 16;
   }

   fclose(infile);
}

/*-----------------------------------------------------------------*/
void step(void) {

   bundle vliw;
   uint32 status;
   uint64 oldip;

   memcpy( &vliw, mem+ip.q, sizeof(bundle) );

   oldip = ip.q;
   if( (status=execute_instruction( &vliw, currslot )) ) {
      printf("error: %s\n", error_strings[status]);
      return;
   }
   else {
      if( currslot==SLOT2 ) {
         currslot = SLOT0;
         if(oldip==ip.q)
            ip.q += 16;
      }
      else {
         currslot <<= 1;
      }
   }
}

/*-----------------------------------------------------------------*/
int main(int argc, char **argv) {

   uint32 i;

   printf("\nIA64 ISA simulator, version 0.1\n");
   printf("Memory limits: 0x%08x-0x%08x\n", 0, MEMSIZE-1);
   printf("Try 'h' for help\n\n");

   for(i=0; i<128; i++) {
      r[i].dat.q = 0; r[i].nat = 0;
   }
   for(i=0; i<128; i++) {
      f[i].mant.q = 0; f[i].exp = 0; f[i].sgn = 0;
   }
   for(i=0; i<64; i++)  p[i] = 0;
   for(i=0; i<128; i++)  ar[i].q = 0;
   for(i=0; i<8; i++)  b[i].q = 0;
   ip.q = currmem = 0;
   cfm.q = 0;
   p[0] = 1;
   currslot=1;
   num_reg_written = 3;
   reg_written[0] = 0 | ALU_REG;
   reg_written[1] = 0 | FPU_REG;
   reg_written[2] = 1 | FPU_REG;

   mem = (uint8 *)calloc( MEMSIZE, sizeof(uint8) );
   if(!mem) {
      printf("memory allocation failed; exiting\n");
      exit(-1);
   }

   while(1) {
      printf("IA64sim>> ");
      gets(cmd);

      nextchar = cmd;
      while(*nextchar) *nextchar++ = tolower(*nextchar);

      if(strchr(cmd,'=')) {
         assign();
      }
      else {
         nextchar = strtok(cmd, tokens);
         if(!nextchar) continue;

         switch( *nextchar ) {
   
            case 'a': assign(); break;
            case 'c': compile(); break;
            case 'h': help(); break;
            case 'i': step(); break;
            case 'p': pred(); break;
            case 'q': free(mem); return 0;
            case 'r': showreg(); break;
            case 'x': showmem(); break;
   
            default:
               printf("  syntax error\n");
         }
      }
      printf("\n");
      r[0].dat.q = 0;
      r[0].nat = 0;
      p[0] = 1;
      ip.q &= ~15;
   }

   return 0;   /* should never get here */
}
