/*--------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Jason Papadopoulos. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors, inconsistencies, and inaccuracies with real-world hardware
implementations of the IA64 instruction set. If you build a mission-
critical system based only on this simulator and something goes wrong,
it's your own damned fault.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.	
       				   --jasonp@glue.umd.edu  12/5/99
--------------------------------------------------------------------*/
/* ia64.h:
   header file for IA64 ISA and runtime
   environment.
*/

#ifndef _IA64_H_
#define _IA64_H_

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define TRUE 1
#define FALSE 0

        /* extract up to 31 bits from the qword x,
           from positions lo to hi inclusive.
           0<=lo,hi<=63 and lo<=hi */

#define BITS(x,lo,hi) (      \
           (uint32)(x##.q >> lo) &   \
	   ( (1<<(hi-lo+1)) - 1 )    \
	   )

#define SWAPI(x,y) {                         \
                uint32 temp;                 \
                temp=(x); (x)=(y); (y)=temp; \
             }

#define MEMSIZE (1<<20)

/*--------------------------- TYPES -------------------------------*/
typedef unsigned long long uint64;
typedef long long int64;
typedef unsigned long uint32;
typedef long int32;
typedef unsigned short uint16;
typedef short int16;
typedef unsigned char uint8;
typedef char int8;

typedef union {
   uint64 q;
   uint32 d[2];
   uint16 w[4];
   uint8 b[8];
} qword;

#define HI(x)  (x##.d[1])
#define LO(x)  (x##.d[0])

typedef struct {
   qword dat;
   uint32 nat;
} alu_reg;

typedef struct {
   qword mant;
   uint32 exp;
   uint32 sgn;
} fpu_reg;

typedef struct {
   qword lo_q;
   qword hi_q;
} bundle;

struct _stack_frame {
   struct _stack_frame *last;
   alu_reg frame[96];
};

typedef struct _stack_frame stack_frame;

/*--------------------------- GLOBALS --------------------------------*/

#define ALU_REG 0x0100                /* register types in the */
#define FPU_REG 0x0200                     /* reg_written list */
#define APP_REG 0x0300
#define PRD_REG 0x0400
#define BR_REG  0x0500
#define PC_REG  0x0600

extern alu_reg r[128];                    /* integer registers */
extern fpu_reg f[128];             /* floating point registers */
extern qword ar[128];                 /* application registers */
extern qword b[8];                         /* branch registers */
extern uint32 p[64];                    /* predicate registers */
extern qword ip;                            /* instruction ptr */
extern qword cfm;                      /* current frame marker */

extern qword slot[3];                 /* instruction slots */
extern stack_frame *last_frame;

extern uint32 num_reg_written;  /* number of registers written so far */
extern uint32 reg_written[128];   /* list of registers written so far */
extern uint8 *mem;                          /* pointer to IA64 memory */

#define AR_PFS 64                       /* previous function state */
#define AR_LC 65                                   /* loop counter */
#define AR_EC 66                                 /* epilog counter */

extern char error_strings[10][20];

enum templates{ mii, miis, misi, misis, mx, mxs,
                mmi=8, mmis, msmi, msmis, mfi, mfis,
                mmf, mmfs, mib, mibs, mbb, mbbs,
                bbb=22, bbbs, mmb, mmbs, mfb=28, mfbs };

/*--------------------------- FUNCTIONS-------------------------------*/

uint32 execute_instruction(bundle *vliw, uint32 mask);

/*-------------------------- ERROR STUFF -----------------------------*/

enum error_codes{
   IA_OK=0,
   INVALID_BUNDLE,
   ILLEGAL_INSTR,
   REGISTER_HAZARD,
   NAT_CONSUMPTION,
   PARSE_ERROR,
   MEMORY_EXCEPTION,
   STACK_FAULT,
   UNIMPLEMENTED
};


#endif /* !_IA64_H_ */
